<?php

namespace Ems\Services;


use Ems\Repository\{ EstruturaRepository, VinculoRepository, ParamOrgRepository };
use Ems\Services\OperadorService;
use Ems\Util\ArrayUtil;
use \Util\Environment;
use Util\Messages;
use Util\DateUtil;
use Setforms2\Shared\ORM\DateTime;

class EstruturaService
{
    protected $estruturaRepository;
    protected $environement;

    /** @var ParamOrgRepository **/
    protected $paramOrgRepository;

    const INDEX_ORDER_ESTRUTURAS_ARRAY = 'NMESTRUTURA';
    
    public const ESTRUTURA_GERENCIAL    = 'NRESTRUTGEREN';
    public const ESTRUTURA_LEGAL        = 'NRESTRUTLEGAL';
    
    public const DEFAULT_ESTRUTURA_PONTO_ACESSO     = self::ESTRUTURA_GERENCIAL;
    public const SECONDARY_ESTRUTURA_PONTO_ACESSO   = self::ESTRUTURA_LEGAL;

    public function __construct(
        OperadorService     $operadorService,
        EstruturaRepository $estruturaRepository,
        VinculoRepository   $vinculoRepository,
        Environment         $environment,
        ParamOrgRepository  $paramOrgRepository
    ) {
        $this->operadorService      = $operadorService;
        $this->estruturaRepository  = $estruturaRepository;
        $this->vinculoRepository    = $vinculoRepository;
        $this->environment          = $environment;
        $this->paramOrgRepository   = $paramOrgRepository;
    }

    /**
     *  Rotina que retorna as estruturas gerenciadas pelo operador
     *
     *  @param string $cdoperador
     *
     *  @return Array $nrEstruturas
     */
    public function getEstruturasGerenciais(
        string $cdOperador
    ) : array {

        $nrEstruturasResp = $this->getEstruturasGerenciaisResp($cdOperador);

        $relacEstruturas = [];

        foreach ($nrEstruturasResp as $nrEstruturaResp) {
            $nrEstruturasInferiores = [];

            $this->getTodasEstruturasInferiores(
                $nrEstruturaResp,
                false,
                $nrEstruturasInferiores
            );

            $relacao = [
                'superior'   => $nrEstruturaResp,
                'inferiores' => $nrEstruturasInferiores
            ];

            array_push($relacEstruturas, $relacao);
        }

        $estruturasFiltradas = $this->
            removerEstruturasSeForemInferiores($relacEstruturas);

        $estruturasOrdenadas = $this->
            ordenarEstruturasPorNumeroInferiores($estruturasFiltradas);

        $estruturas = $this->
            planificarRelacEstruturas($estruturasOrdenadas);

        return $estruturas;
    }

    public function insertEstruturaTodas(
        string $cdOperador,
        array &$estruturas
    ) {
        $isOperadorDP = $this->operadorService->operadorIsDP($cdOperador);

        if ($isOperadorDP) {
            $estruturaTodas = [
                'IDTIPOESTRUTURA'   => null,
                'NMESTRUTURA'       => '-- Todas --',
                'NRESTRUTURAM'      => '-1',
            ];

            array_unshift($estruturas, $estruturaTodas);
        }
    }

    // public function getTodasEstruturasInferiores(
    //     int $nrEstrutura,
    //     $selfInclude = false
    // ) {
    //     $nrEstruturasInferiores = [];

    //     $this->getEstruturasInferioresRec(
    //         $nrEstrutura, $nrEstruturasInferiores, $selfInclude
    //     );

    //     return $nrEstruturasInferiores;
    // }

    public function getTodasEstruturasInferiores(
        int $nrEstrutura,
        $selfInclude = false,
        &$children = []
    ) {
        if (in_array($nrEstrutura, $children)) {
            return;
        }

        if ($selfInclude) {
            array_push($children, $nrEstrutura);
        }

        // Todas as estruturas que são *DIRETAMENTE* inferiores à estrutua passada
        $directChildren = $this->getNrsEstruturasInferiores($nrEstrutura);

        foreach ($directChildren as $child) {
            $this->getTodasEstruturasInferiores($child, true, $children);
        }

        return $children;
    }


    /**
     *  Rotina que retorna as estruturas imediatamente abaixo ao operador. Caso o operador seja do perfil de acesso administrador do sistema, serão retornadas todas as estruturas gerenciaise legais ativas
     *
     *  @param string $cdoperador
     *
     *  @return Array $nrEstruturas
     */
    public function getEstruturasGerenciaisResp(
        string $cdOperador
    ) : array {

        $nrOrg = (int) $this->environment->getNrOrg();

        // Se for DP, retorna todas as estruturas gerenciais da organização. Caso contrário, são retornadas as estruturas imediatamente abaixo ao operador
        $isOperadorDP = $this->operadorService->operadorIsDP($cdOperador);

        $estruturas = [];

        if ($isOperadorDP) {
            $estruturas = $this->estruturaRepository->getEstruturasAcessiveisDP($nrOrg);
        } else {

            $params = [
                'P_NRORG'      => $nrOrg,
                'P_CDOPERADOR' => $cdOperador
            ];

            $estruturas = $this->estruturaRepository->getEstruturasGerenciaisResp($params);
        }


        $nrEstruturas = array_map(function(array $estrutura) {
            $nrEstrutura = (int) $estrutura['NRESTRUTURAM'];

            return $nrEstrutura;
        }, $estruturas);

        return $nrEstruturas;
    }

    /** Retorna um array(int) com os NRESTRUTURAM das estrutuas inferiores à
     * estrutura informada.
    */
    public function getNrsEstruturasInferiores(
        int $nrEstrutura
    ) {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'       => $nrOrg,
            'P_NRESTRUTURA' => $nrEstrutura
        ];

        $estruturas = $this->estruturaRepository->getEstruturasInferiores($params);

        $nrEstruturas = array_map(function(array $estrutura) {
            $nrEstrutura = (int) $estrutura['NRESTRUTURAMREL'];

            return $nrEstrutura;
        }, $estruturas);

        return $nrEstruturas;
    }

    public function getNomeETipoEstrutura(
        int $nrEstrutura
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'       => $nrOrg,
            'P_NRESTRUTURA' => $nrEstrutura
        ];

        $result = $this->estruturaRepository->getNomeETipoEstrutura($params);

        return $result;
    }

    public function removerEstruturasSeForemInferiores($relacEstruturas) {
        foreach ($relacEstruturas as $key => $relacao) {
            $superior = $relacao['superior'];

            foreach($relacEstruturas as $relac) {
                if ($relac['superior'] !== $superior) {
                    $isInferior = in_array($superior, $relac['inferiores']);

                    if ($isInferior) {
                        unset($relacEstruturas[$key]);
                    }
                }
            }
        }

        return $relacEstruturas;
    }

    public function planificarRelacEstruturas($relacEstruturas) {
        $estruturas = [];

        foreach ($relacEstruturas as $relacao) {
            array_push($estruturas, $relacao['superior']);
            $estruturas = array_merge($estruturas, $relacao['inferiores']);
        }

        return $estruturas;
    }

    public function ordenarEstruturasPorNumeroInferiores($relacEstruturas) {
        usort($relacEstruturas, function($a, $b) {
            $qntInferioresA = count($a['inferiores']);
            $qntInferioresB = count($b['inferiores']);

            $diferenca = $qntInferioresB - $qntInferioresA;

            return $diferenca;
        });

        return $relacEstruturas;
    }

    public function getDadosEstrutura(
        string $nrestruturam,
        string $nrorg = null
    ) {
        if (empty($nrorg)) {
            $nrorg = $this->environment->getNrorg();
        }

        $estrutura = $this->estruturaRepository->getEstruturaMestreHistorico($nrestruturam, $nrorg);

        if (empty($estrutura)) {
            throw new \Exception(Messages::ESTRUT_NOT_FOUND);
        }

        return $estrutura;
    }

    public function retornaDadosResponsavelInformal($nrorg, $nrestrutgeren, $dtmescompetenc){
        return $this->estruturaRepository->retornaDadosResponsavelInformal($nrorg, $nrestrutgeren, $dtmescompetenc);
    }

    public function getEstrutura(
        $nmTipoEstrutura = null
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();
        $nrOrgPadrao = (int) $this->environment->getNrOrgPadrao();

        $dtMesCompetencia = DateUtil::getPrimeiroDiaDoMes(new DateTime());
        $dtMesCompetencia = $dtMesCompetencia->format('d/m/Y');

        $nrTipoEstrutura = null;

        if ($nmTipoEstrutura !== null) {
            if (is_numeric($nmTipoEstrutura)) {
                $nrTipoEstrutura = $nmTipoEstrutura;
            } else {
                $paramOrg = $this->paramOrgRepository->getParamOrg($nrOrg);

                if ($nmTipoEstrutura === 'LEGAL') {
                    $nrTipoEstrutura = $paramOrg['NRTPESTRUTLEGAL'];
                } else if ($nmTipoEstrutura === 'GERENCIAL') {
                    $nrTipoEstrutura = $paramOrg['NRTPESTRUTGEREN'];
                } else if ($nmTipoEstrutura === 'SINDICAL') {
                    $nrTipoEstrutura = 10;
                }
            }
        }

        $params = [
            'P_NRORG'           => $nrOrg,
            'P_NRORG_PADRAO'    => $nrOrgPadrao,
            'P_DTMESCOMPETENC'  => $dtMesCompetencia,
            'P_NRTIPOESTRUTURA' => $nrTipoEstrutura,
        ];

        $result = $this->estruturaRepository->getEstrutura($params);

        return $result;
    }

    public function getEstruturaAdicional(
        $nmTipoEstrutura = null
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();
        $nrOrgPadrao = (int) $this->environment->getNrOrgPadrao();

        $dtMesCompetencia = DateUtil::getPrimeiroDiaDoMes(new DateTime());
        $dtMesCompetencia = $dtMesCompetencia->format('d/m/Y');

        $nrTipoEstrutura = null;

        $params = [
            'P_NRORG'           => $nrOrg,
            'P_NRORG_PADRAO'    => $nrOrgPadrao,
            'P_DTMESCOMPETENC'  => $dtMesCompetencia,
        ];
        
        $result = $this->estruturaRepository->getEstruturaAdicional($params);

        return $result;
    }

    public function getAniversariantesEstrutura(
        int $nrEstrutura
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();

        $nrEstruturasInferiores = $this->getTodasEstruturasInferiores($nrEstrutura, true);
        $nrEstruturasInferiores = ArrayUtil::toStringSQL($nrEstruturasInferiores);

        $dtMesCompetencia = date('01/m/Y');

        $params = [
            'P_NRORG' => $nrOrg,
            'P_NRESTRUTURAS' => $nrEstruturasInferiores,
            'P_DTMESCOMPETENC' => $dtMesCompetencia,
        ];

        $aniversariantes = $this->estruturaRepository->getAniversariantesEstrutura($params);

        usort($aniversariantes, function($a, $b) {
            return self::sortAniver($a, $b);
        });

        return $aniversariantes;
    }

    public static function sortAniver(
        array $a,
        array $b
    ): int {
        $currentDate = DateUtil::getDataAtual()->format(DateUtil::FORMATO_BRASILEIRO);
        $currentDay = intval(substr($currentDate, 0, 2));

        $dayA = intval(substr($a['DTANIVERSARIO'], 0, 2));
        $dayB = intval(substr($b['DTANIVERSARIO'], 0, 2));

        if ($currentDay === $dayA) {
            return -1;
        }

        if ($currentDay === $dayB) {
            return 1;
        }

        return $dayA - $dayB;
    }
}
